/*jslint node: true */
"use strict";

var util=require("util");
var semver=require("semver");

var AHTDPACL=function AHTDPACL(api)
{
	var ACLAllowMap={};
	var ACLRoleInheritanceMap={};
	var config;
	var logger;

	var getAvailableActions=function AHTDPACLGetAvailableActions(api)
	{
		if("actions" in api)
		{
			if(typeof(api.actions.versions)==="object")
			{
				return api.actions.versions;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}
	
	
	var normaliseActionVersion=function AHTDPACLNormaliseActionVersion(actionVersion)
	{
		if(!semver.valid(actionVersion))
		{
			actionVersion=actionVersion.toString();

			if(actionVersion.match(/^([0-9]+)$/))
			{
				actionVersion=actionVersion+".0.0";
			}
			else if(actionVersion.match(/^([0-9]+)\.([0-9]+)$/))
			{
				actionVersion=actionVersion+".0";
			}

			// Last ditch fix
			if(!semver.valid(actionVersion))
			{
				actionVersion="1.0.0";
			}
		}
		
		return actionVersion;
	};
	
	var normaliseActionName=function AHTDPACLNormaliseActionName(actionName)
	{
		return actionName.toString().trim();
	};
	
	var mapAllowedRolesForActionVersion=function AHTDPACLMapAllowedRolesForActionVersion(actionName, actionVersion, rules, ruleProcessingOrder)
	{
		var allowedRoles=[];
		var i;
		var j;
		var k;
		var actionNameRegexp

		if(ruleProcessingOrder.replace(" ", "").toLowerCase()==="allow,deny")
		{
			var ruleType=["allow","deny"];
		}
		else
		{
			var ruleType=["deny","allow"];
		}
		
		var normalisedActionVersion=normaliseActionVersion(actionVersion);
		
		for(j in ruleType)
		{
			for(i in rules)
			{
				if(typeof(rules[i])==="object")
				{
					for(k in rules[i][ruleType[j]])
					{
						var r=rules[i][ruleType[j]][k].split(":");

						var ruleAction=r[0];
						var ruleVersion=r[1] || ">=0.0";

						if(typeof(ruleAction)==="string")
						{
							actionNameRegexp=new RegExp("^"+ruleAction.replace("*", ".*", "g")+"$");				
						}
						else if(util.isRegExp(ruleAction))
						{
							actionNameRegexp=ruleAction;
						}
						
						// We'll add an entry in the ACLAllowMap for all actions/versions to make checling them simpler
						if(typeof(ACLAllowMap[actionName])==="undefined")
						{
							ACLAllowMap[actionName]=[];
						}
						if(typeof(ACLAllowMap[actionName][normalisedActionVersion])==="undefined")
						{
							// NOTE: allow is calculated (i.e. a deny rule after an allow will result in removal of the allow rule)
							// deny is simply a collection of deny's and is used to deny access to actions if e.g. the actual role has a deny rule for action A but one of the roles the role inherits from has an allow for it
							ACLAllowMap[actionName][normalisedActionVersion]=
							{
								allow:[],
								deny:[]
							};
						}
						
						if(actionName.match(actionNameRegexp))
						{
							if(semver.satisfies(normalisedActionVersion, ruleVersion))
							{	
								// We have a match so we'll add/remove rules
								if(ruleType[j]==="allow") // Allow - add perm for action/version/user
								{
									ACLAllowMap[actionName][normalisedActionVersion].allow.push(i);
								}
								else // deny - remove perm for action/version/user
								{
									ACLAllowMap[actionName][normalisedActionVersion].allow.splice(i);
									ACLAllowMap[actionName][normalisedActionVersion].deny.push(i);
								}
							}
						}	
					}
				}
			}
		}
		
		return allowedRoles;
	}
	
	
	var mapRoleInheritance=function AHTDPACLMapRoleInheritance(rules)
	{
		var i,j;
		var roleInheritanceMap={};
		
		for(i in rules)
		{
			if(typeof(rules[i])==="object")
			{
				if(typeof(rules[i].inheritsFromRoles)!=="undefined")
				{
					if(typeof(rules[i].inheritsFromRoles)!=="undefined")
					{
						if(typeof(roleInheritanceMap[i])!=="object")
						{
							roleInheritanceMap[i]=[];
						}
					}
					
					if(typeof(rules[i].inheritsFromRoles)==="string")
					{
						roleInheritanceMap[i].push(rules[i].inheritsFromRoles)
					}
					else if(typeof(rules[i].inheritsFromRoles)==="object")
					{
						for(j in rules[i].inheritsFromRoles)
						{
							roleInheritanceMap[i].push(rules[i].inheritsFromRoles[j]);
						}
					}
				}
			}
		}
				
		return roleInheritanceMap;
	}
	

	var mapRoleInheritanceOntoACL=function()
	{
		if(typeof(ACLAllowMap)==="object")
		{
			var a,v,r,ir;
			for(a in ACLAllowMap)
			{
				for(v in ACLAllowMap[a])
				{
					// If the existing, non-inheritance-based ACLAllowMap has one of the roles in it which this role inherits from, add this role
					for(r in ACLRoleInheritanceMap)
					{
						for(ir in ACLRoleInheritanceMap[r])
						{
							if(typeof(ACLAllowMap[a][v].allow)==="object")
							{
								// If allow map contains this inherited role...and main role is not yet present
								if(ACLAllowMap[a][v].allow.indexOf(ACLRoleInheritanceMap[r][ir])!==-1 && ACLAllowMap[a][v].allow.indexOf(r)===-1)		
								{
									// push the role onto the allow map
									ACLAllowMap[a][v].allow.push(r);
								}
							}

							if(typeof(ACLAllowMap[a][v].deny)==="object")
							{
								// If allow map contains this inherited role...and main role is not yet present
								if(ACLAllowMap[a][v].deny.indexOf(ACLRoleInheritanceMap[r][ir])!==-1 && ACLAllowMap[a][v].deny.indexOf(r)===-1)		
								{
									// push the role onto the allow map
									ACLAllowMap[a][v].deny.push(r);
								}
							}
						}
						
					}
				}
			}

			// Would be nice to add some more error checking
			return true;
		}
		else
		{
			return false;
		}
	}


		
	var init=function AHTDPACLInit(api)
	{
		var actions;
		var actionName;
		var actionVersion;
		
		// Public vars
		logger=api.log;
		
		logger("ah-tdp-acl-plugin: Beginning initialisation","debug");

		if(typeof(api.config.AHTDPACLPlugin)==="object")
		{
			config=api.config.AHTDPACLPlugin;

			if(actions=getAvailableActions(api))
			{
				logger("ah-tdp-acl-plugin: Got available API actions","debug");

				for(actionName in actions)
				{
					for(actionVersion in actions[actionName])
					{			
						if(mapAllowedRolesForActionVersion(actionName, actions[actionName][actionVersion], config.rules, config.ruleProcessingOrder)===false && config.exitOnRoleProcessingError!==false)
						{
							logger("ah-tdp-acl-plugin: mapAllowedRolesForActionVersion returned false (failure) for action "+actionName+", version "+actionVersion, "error");
							return false;
						}
					}
				}
				
				if(ACLRoleInheritanceMap=mapRoleInheritance(config.rules))
				{
					return mapRoleInheritanceOntoACL();
				}
				else
				{
					return false;
				}
			}
			else
			{
				logger("ah-tdp-acl-plugin: Couldn't find actions from api object","error");
				return false;
			}

		}
		else
		{
			logger("ah-tdp-acl-plugin: Couldn't find config at api.config.AHTDPACLPlugin. Make sure the TDP AH ACL config file is in <root>/config/plugins","error");
			return false;
		}
	};



// Public methods	
	this.normaliseActionVersion=function normaliseActionVersionFn(actionVersion)
	{

// could this be async?		
		return normaliseActionVersion(actionVersion);
	}
	
	this.normaliseActionName=function normaliseActionNameFn(actionName)
	{
// could this be async?		
		return normaliseActionName(actionName);
	}
	
	this.roleHasPermissionsOnAction=function roleHasPermissionsOnActionFn(role, actionName, actionVersion, callback)
	{
		var i;
		var normalisedActionName=this.normaliseActionName(actionName);
		var normalisedActionVersion=this.normaliseActionVersion(actionVersion);

		if(ACLAllowMap[normalisedActionName])
		{
			if(ACLAllowMap[normalisedActionName][normalisedActionVersion])
			{
				if(ACLAllowMap[normalisedActionName][normalisedActionVersion].allow.indexOf(role)!=-1 && ACLAllowMap[normalisedActionName][normalisedActionVersion].deny.indexOf(role)===-1)
				{
					logger("ah-tdp-acl-plugin: Action "+normalisedActionName+" (@v. "+normalisedActionVersion+") allowed for role "+role, "debug");
					return callback(true);
				}
				else 
				{
					logger("ah-tdp-acl-plugin: Action "+normalisedActionName+" (@v. "+normalisedActionVersion+") not allowed for role "+role, "debug");
					return callback(false);
				}
			}
		}

		// Catchall exit point - we should never get here in practice
		return callback(false);
	};
	
	
	
	this.getUserRole=function getUserRoleFn(connection, callback)
	{
		var err=null;
		var role=null;
		
		if(typeof(connection)==="object")
		{
			connection[config.sessionDataObjectNameOnConnection]=connection[config.sessionDataObjectNameOnConnection] || {};

			if(typeof(connection[config.sessionDataObjectNameOnConnection])==="object")
			{
				if(typeof(connection[config.sessionDataObjectNameOnConnection][config.roleStringNameOnSessionDataObject])==="string")
				{
					role=connection[config.sessionDataObjectNameOnConnection][config.roleStringNameOnSessionDataObject];
				}
				else if(config.defaultRole)
				{
					role=config.defaultRole;

					// We'll also set the default role on the connection object as this may be session'd
					connection[config.sessionDataObjectNameOnConnection][config.roleStringNameOnSessionDataObject]=role;
				}
				else
				{
					err="No role found";
				}
			}
			else
			{
				err="connection."+config.sessionDataObjectNameOnConnection+" is not an object";
			}
		}
		else
		{
			err="connection is not an object";
		}

		return callback(err, role);
	}
	
	if(typeof(api)==="object")
	{
		init(api);
	}
	
};


module.exports=AHTDPACL;